@@ -1,23 +1,21 @@
 """Calculator for Cluster Expansion."""
 import sys
 import contextlib
-from typing import Dict, Optional, TextIO, Union, List, Sequence
+from typing import Dict, Optional, TextIO, Union, List
 import numpy as np
 from ase import Atoms
 from ase.calculators.calculator import Calculator
-from clease.datastructures import SystemChange
+from clease_cxx import PyCEUpdater
+from clease.datastructures import SystemChange, SystemChanges
 from clease.corr_func import CorrFunction
 from clease.settings import ClusterExpansionSettings
-from clease_cxx import PyCEUpdater
 
 
 class MovedIgnoredAtomError(Exception):
     """Raised when ignored atoms is moved."""
 
 
-# pylint: disable=too-few-public-methods
 class KeepChanges:
-
     def __init__(self):
         self.keep_changes = True
 
@@ -40,14 +38,16 @@ class Clease(Calculator):
         file. Use '-' for stdout.
     """
 
-    name = 'CLEASE'
-    implemented_properties = ['energy']
+    name = "CLEASE"
+    implemented_properties = ["energy"]
 
-    def __init__(self,
-                 settings: ClusterExpansionSettings,
-                 eci: Dict[str, float],
-                 init_cf: Optional[Dict[str, float]] = None,
-                 logfile: Union[TextIO, str, None] = None) -> None:
+    def __init__(
+        self,
+        settings: ClusterExpansionSettings,
+        eci: Dict[str, float],
+        init_cf: Optional[Dict[str, float]] = None,
+        logfile: Union[TextIO, str, None] = None,
+    ) -> None:
         Calculator.__init__(self)
 
         if not isinstance(settings, ClusterExpansionSettings):
@@ -75,10 +75,11 @@ class Clease(Calculator):
 
         # logfile
         if isinstance(logfile, str):
-            if logfile == '-':
+            if logfile == "-":
                 logfile = sys.stdout
             else:
-                logfile = open(logfile, 'a')
+                # pylint: disable=consider-using-with
+                logfile = open(logfile, "a")
         self.logfile = logfile
 
         # reference atoms for calculating the cf and energy for new atoms
@@ -89,44 +90,11 @@ class Clease(Calculator):
         # C++ updater initialized when atoms are set
         self.updater = None
 
-    def _set_norm_factors(self) -> None:
-        """Set normalization factor for each cluster.
-
-        The normalization factor only kicks in when the cell is too small and
-        thus, include self-interactions. This methods corrects the impact of
-        self-interactions.
-        """
-        symm_group = np.zeros(len(self.settings.atoms), dtype=np.uint8)
-        for num, group in enumerate(self.settings.index_by_sublattice):
-            symm_group[group] = num
-
-        cluster_list = self.settings.cluster_list
-        for cluster in cluster_list:
-            fig_keys = list(set(cluster.get_all_figure_keys()))
-            num_occ = {}
-            for key in fig_keys:
-                num_occ[key] = cluster_list.num_occ_figure(key, cluster.name, symm_group,
-                                                           self.settings.trans_matrix)
-            num_fig_occ = cluster.num_fig_occurences
-            norm_factors = {}
-            for key in fig_keys:
-                tot_num = num_occ[key]
-                num_unique = len(set(key.split("-")))
-                norm_factors[key] = \
-                    float(tot_num)/(num_unique*num_fig_occ[key])
-
-            norm_factor_list = []
-            for fig in cluster.indices:
-                key = cluster.get_figure_key(fig)
-                norm_factor_list.append(norm_factors[key])
-            cluster.info['normalization_factor'] = norm_factor_list
-
     def set_atoms(self, atoms: Atoms) -> None:
         self.atoms = atoms
 
         if self.init_cf is None:
-            self.init_cf = \
-                self.corrFunc.get_cf_by_names(self.atoms, self.cf_names)
+            self.init_cf = self.corrFunc.get_cf_by_names(self.atoms, self.cf_names)
 
         if len(self.settings.atoms) != len(atoms):
             msg = "Passed Atoms object and settings.atoms should have same "
@@ -144,13 +112,15 @@ class Clease(Calculator):
         self.is_backround_index = np.zeros(len(atoms), dtype=np.uint8)
         self.is_backround_index[self.settings.background_indices] = 1
 
-        self._set_norm_factors()
-        self.updater = PyCEUpdater(self.atoms, self.settings, self.init_cf, self.eci,
-                                   self.settings.cluster_list)
+        self.updater = PyCEUpdater(
+            self.atoms,
+            self.settings,
+            self.init_cf,
+            self.eci,
+            self.settings.cluster_list,
+        )
 
-    def get_energy_given_change(self,
-                                system_changes: Sequence[SystemChange],
-                                keep_changes=False) -> float:
+    def get_energy_given_change(self, system_changes: SystemChanges, keep_changes=False) -> float:
         """
         Calculate the energy when the change is known. No checking will be
         performed.
@@ -172,16 +142,32 @@ class Clease(Calculator):
         res = super().check_state(atoms)
         syst_ch = self.indices_of_changed_atoms
 
-        if syst_ch and 'energy' not in res:
-            res.append('energy')
+        if syst_ch and "energy" not in res:
+            res.append("energy")
         return res
 
     def reset(self):
         self.results = {}
 
+    def calculate_cf_from_scratch(self) -> Dict[str, float]:
+        """Calculate correlation functions from scratch."""
+        return self.updater.calculate_cf_from_scratch(self.atoms, self.cf_names)
+
+    def calculate_energy_from_scratch(self) -> float:
+        """Calculate the correlation functions, and hence also the energy,
+        from scratch.
+        """
+        self.calculate_cf_from_scratch()
+        self.energy = self.updater.get_energy()
+        return self.energy
+
     # pylint: disable=signature-differs
-    def calculate(self, atoms: Atoms, properties: Union[List[str], str],
-                  system_changes: Sequence[SystemChange]) -> float:
+    def calculate(
+        self,
+        atoms: Atoms,
+        properties: Union[List[str], str],
+        system_changes: SystemChanges,
+    ) -> float:
         """Calculate the energy of the passed Atoms object.
 
         If accept=True, the most recently used atoms object is used as a
@@ -219,7 +205,6 @@ class Clease(Calculator):
         """
         self.updater.undo_changes()
         self.energy = self.updater.get_energy()
-        self.results['energy'] = self.energy
 
     def update_energy(self) -> None:
         """Update correlation function and get new energy."""
@@ -228,19 +213,18 @@ class Clease(Calculator):
 
     @property
     def energy(self):
-        return self.results.get('energy', None)
+        return self.results.get("energy", None)
 
     @energy.setter
     def energy(self, value):
-        self.results['energy'] = value
+        self.results["energy"] = value
 
     @property
     def indices_of_changed_atoms(self) -> List[int]:
         """Return the indices of atoms that have been changed."""
         changed = self.updater.get_changed_sites(self.atoms)
         for index in changed:
-            if self.is_backround_index[index] and \
-                    not self.settings.include_background_atoms:
+            if self.is_backround_index[index] and not self.settings.include_background_atoms:
                 msg = f"Atom with index {index} is a background atom."
                 raise MovedIgnoredAtomError(msg)
 
@@ -250,7 +234,7 @@ class Clease(Calculator):
         """Return the correlation functions as a dict"""
         return self.updater.get_cf()
 
-    def update_cf(self, system_changes: Sequence[SystemChange] = None) -> None:
+    def update_cf(self, system_changes: SystemChanges = None) -> None:
         """Update correlation function based on the reference value.
 
         :param system_changes: List of system changes. For example, if the
@@ -258,15 +242,20 @@ class Clease(Calculator):
             system_change = [(23, Mg, Al)]. If an Mg atom occupying the atomic
             index 26 is swapped with an Al atom occupying the atomic index 12,
             system_change = [(26, Mg, Al), (12, Al, Mg)]
+            If system_changes is None, the correlation functions are recalculated.
+            Default: None.
         """
         if system_changes is None:
             swapped_indices = self.indices_of_changed_atoms
             symbols = self.updater.get_symbols()
             system_changes = [
-                SystemChange(index=idx,
-                             old_symb=symbols[idx],
-                             new_symb=self.atoms[idx].symbol,
-                             name='internal_symbol_change') for idx in swapped_indices
+                SystemChange(
+                    index=idx,
+                    old_symb=symbols[idx],
+                    new_symb=self.atoms[idx].symbol,
+                    name="internal_symbol_change",
+                )
+                for idx in swapped_indices
             ]
         for change in system_changes:
             self.updater.update_cf(change)
@@ -307,7 +296,7 @@ class Clease(Calculator):
         """
 
     @contextlib.contextmanager
-    def with_system_changes(self, system_changes: Sequence[SystemChange]):
+    def with_system_changes(self, system_changes: SystemChanges):
         """
         : param system_changes: List of system changes. For example, if the
             occupation of the atomic index 23 is changed from Mg to Al,
@@ -317,28 +306,30 @@ class Clease(Calculator):
         """
         keeper = KeepChanges()  # We need an object we can mutate to flag for cleanup
         try:
-            self.reset()
             # Apply the updates
-            self.update_cf(system_changes)
-            self.energy = self.updater.get_energy()
+            self.apply_system_changes(system_changes)
             yield keeper
         finally:
             if keeper.keep_changes:
                 # Keep changes
-                self.clear_history()
+                self.keep_system_changes()
             else:
                 # Revert changes
-                inverted_changes = self._get_inverted_changes(system_changes)
-                self.restore()  # Also restores results
-                for idx, _, symb, _ in inverted_changes:
-                    self.atoms.symbols[idx] = symb
+                self.undo_system_changes()
 
-    @staticmethod
-    def _get_inverted_changes(system_changes: Sequence[SystemChange]):
-        """Invert system changes by doing old_symbs -> new_symbs"""
-        return [
-            SystemChange(index=change.index,
-                         old_symb=change.new_symb,
-                         new_symb=change.old_symb,
-                         name=change.name) for change in system_changes
-        ]
+    def apply_system_changes(self, system_changes: SystemChanges) -> None:
+        """Apply a set of changes to the calculator, and evaluate the energy"""
+        self.reset()
+        # Apply the updates
+        self.update_cf(system_changes)
+        self.energy = self.updater.get_energy()
+
+    def undo_system_changes(self) -> None:
+        """Revert a set of changes. The changes passed in should be the original
+        sequence of system changes which were applied. Restores the original results."""
+        self.restore()  # Also restores results
+
+    def keep_system_changes(self) -> None:
+        """A set of system changes are to be kept. Perform necessary actions to prepare
+        for a new evaluation."""
+        self.clear_history()
