@@ -1,19 +1,18 @@
 """Monte Carlo method for ase."""
-from typing import Tuple, Dict, Union
+from typing import Dict, Union, Iterator
 import sys
 import datetime
 import time
 import logging
 import random
+import math
 from collections import Counter
-import numpy as np
 from ase import Atoms
 from ase.units import kB
 from clease.version import __version__
-from clease.datastructures import SystemChange, SystemChanges
-from .mc_evaluator import MCEvaluator
+from clease.datastructures import SystemChanges, MCStep
+from .mc_evaluator import CEMCEvaluator, MCEvaluator
 from .base import BaseMC
-from .exponential_filter import ExponentialFilter
 from .averager import Averager
 from .bias_potential import BiasPotential
 from .observers import MCObserver
@@ -36,18 +35,21 @@ class Montecarlo(BaseMC):
             :class:`~clease.montecarlo.trial_move_generator.TrialMoveGenerator`
             object that produces trial moves. Defaults to None.
     """
+
     NAME = "MonteCarlo"
 
-    def __init__(self,
-                 system: Union[Atoms, MCEvaluator],
-                 temp: float,
-                 generator: TrialMoveGenerator = None):
+    def __init__(
+        self,
+        system: Union[Atoms, MCEvaluator],
+        temp: float,
+        generator: TrialMoveGenerator = None,
+    ):
 
         super().__init__(system)
         self.T = temp
 
         if generator is None:
-            self.generator = RandomSwap(self.atoms)
+            self.generator = _make_default_swap_generator(self.evaluator)
         else:
             self.generator = generator
 
@@ -70,13 +72,9 @@ class Montecarlo(BaseMC):
 
         self.quit = False
 
-        self.filter = ExponentialFilter(min_time=0.2 * len(self.atoms),
-                                        max_time=20 * len(self.atoms),
-                                        n_subfilters=10)
-
     def update_current_energy(self):
         self.current_energy = self.evaluator.get_energy()
-        logger.debug('Updating current energy to %s', self.current_energy)
+        logger.debug("Updating current energy to %s", self.current_energy)
         self.evaluator.reset()
 
     def _initialize_energies(self) -> None:
@@ -94,12 +92,11 @@ class Montecarlo(BaseMC):
 
     def reset(self):
         """Reset all member variables to their original values."""
-        logger.debug('Resetting.')
+        logger.debug("Resetting.")
         for _, obs in self.observers:
             obs.reset()
 
         self.evaluator.reset()
-        self.filter.reset()
         self._reset_internal_counters()
 
         if self.mean_energy is not None:
@@ -143,7 +140,7 @@ class Montecarlo(BaseMC):
 
     def initialize_run(self):
         """Prepare MC object for a new run."""
-        logger.debug('Initializing run')
+        logger.debug("Initializing run")
         self.generator.initialize(self.atoms)
 
         # Ensure the evaluator is properly synchronized.
@@ -155,7 +152,7 @@ class Montecarlo(BaseMC):
         # Reset the internal step counters
         self._reset_internal_counters()
 
-    def run(self, steps: int = 100):
+    def run(self, steps: int = 100) -> None:
         """Run Monte Carlo simulation.
 
         Parameters:
@@ -164,41 +161,85 @@ class Montecarlo(BaseMC):
             Number of steps in the MC simulation
         """
 
-        logger.info('Starting MC run with %d steps.', steps)
-        # Check the number of different elements are correct to avoid
-        # infinite loops
-        self.initialize_run()
+        # Construct the iterator, make the preparations for starting the run
+        mc_iter = self.irun(steps)
 
-        start = time.time()
+        start = time.perf_counter()
         prev = self.current_step
         info_enabled = logger.isEnabledFor(logging.INFO)  # Do we emit INFO logs?
-        while self.current_step < steps:
-            E, _ = self._mc_step()
-
-            self.mean_energy += E
-            self.energy_squared += E**2
-
+        for _ in mc_iter:
             # We only want to do this calculation if logging is enabled for INFO.
-            if info_enabled and time.time() - start > self.status_every_sec:
+            if info_enabled and time.perf_counter() - start > self.status_every_sec:
                 ms_per_step = 1000.0 * self.status_every_sec / (self.current_step - prev)
                 accept_rate = self.num_accepted / self.current_step
-                logger.info("%d of %d steps. %.2f ms per step. Acceptance rate: %.2f",
-                            self.current_step, steps, ms_per_step, accept_rate)
+                logger.info(
+                    "%d of %d steps. %.2f ms per step. Acceptance rate: %.2f",
+                    self.current_step,
+                    steps,
+                    ms_per_step,
+                    accept_rate,
+                )
                 prev = self.current_step
-                start = time.time()
+                start = time.perf_counter()
+
+    def irun(self, steps: int) -> Iterator[MCStep]:
+        """Run Monte Carlo simulation as an iterator.
+        Can be used to inspect the MC after each step, for example,
+        to print the energy every 5 steps, one could do:
+
+        # doctest: +SKIP
+        >>> mc = Montecarlo(...)
+        >>> for mc_step in mc.irun(500):
+        ...     if mc_step.step % 5 == 0:
+        ...         print(f"Current energy: {mc_step.energy:.2f} eV")
+
+        The iterator yields the current energy, as well as whether the last move
+        was accepted.
+
+        Parameters:
+
+        steps: int
+            Number of steps in the MC simulation
+        """
+        logger.info("Starting MC run with %d steps.", steps)
+
+        # We first ensure the MC object is initialized, and then we
+        # construct the iterator.
+        # This ensures everything is prepared before the first step
+        # is consumed from the iterator.
+        self.initialize_run()
+
+        # Now we create the iterator
+        return self._irun(steps)
+
+    def _irun(self, steps: int) -> Iterator[MCStep]:
+        """Create the MC iterator"""
+        # Offset range by 1, so that we start with current step = 1
+        for _ in range(steps):
+            step = self._mc_step()
+
+            en = step.energy
+            self.mean_energy += en
+            # E * E is slightly faster than E ** 2
+            self.energy_squared += en * en
 
             if self.quit:
-                logger.warning('Quit signal detected. Breaking the MC loop.')
+                # Breaking out will cause the iterator to end.
+                logger.warning("Quit signal detected. Breaking the MC loop.")
                 break
-
-        logger.info("Reached maximum number of steps (%d mc steps)", steps)
+            # Yield, allow inspection before continuing.
+            yield step
+        else:
+            # Loop was not broken, we reached the max number of steps
+            logger.info("Reached maximum number of steps (%d mc steps)", steps)
+        logger.debug("Reached end of MC iterator.")
 
     @property
     def meta_info(self):
         """Return dict with meta info."""
         ts = time.time()
-        st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
-        clease_version = __version__
+        st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")
+        clease_version = str(__version__)
         v_info = sys.version_info
         meta_info = {
             "timestamp": st,
@@ -240,7 +281,7 @@ class Montecarlo(BaseMC):
         system_changes: list
             List with system changes
         """
-        logger.debug('Applying changes to the system.')
+        logger.debug("Applying changes to the system.")
         self.evaluator.apply_system_changes(system_changes)
         # Evaluate the energy quantity after applying the changes to the system.
         self.new_energy = self.evaluator.get_energy(applied_changes=system_changes)
@@ -249,17 +290,19 @@ class Montecarlo(BaseMC):
         # already been introduced to the system
         for bias in self.bias_potentials:
             self.new_energy += bias(system_changes)
-        logger.debug('Current energy: %.3f eV, new energy: %.3f eV', self.current_energy,
-                     self.new_energy)
+        logger.debug(
+            "Current energy: %.3f eV, new energy: %.3f eV",
+            self.current_energy,
+            self.new_energy,
+        )
         accept = self._do_accept(self.current_energy, self.new_energy)
-        logger.debug('Change accepted? %s', accept)
+        logger.debug("Change accepted? %s", accept)
 
         if accept:
             # Changes accepted, finalize evaluator.
             self.evaluator.keep_system_changes(system_changes)
         else:
             # Undo changes
-            logger.debug('Change rejected, undoing system changes.')
             self.evaluator.undo_system_changes(system_changes)
 
         return accept
@@ -279,40 +322,31 @@ class Montecarlo(BaseMC):
             return True
         kT = self.T * kB
         energy_diff = new_energy - current_energy
-        probability = np.exp(-energy_diff / kT)
-        logger.debug('Energy difference: %.3e. Calculated probability: %.3f', energy_diff,
-                     probability)
+        probability = math.exp(-energy_diff / kT)
+        logger.debug(
+            "Energy difference: %.3e. Calculated probability: %.3f",
+            energy_diff,
+            probability,
+        )
         return random.random() <= probability
 
-    def _move_accepted(self, system_changes: SystemChanges) -> SystemChanges:
-        logger.debug('Move accepted, updating things')
+    def _move_accepted(self, system_changes: SystemChanges) -> None:
+        logger.debug("Move accepted, updating things")
         self.num_accepted += 1
         self.generator.on_move_accepted(system_changes)
         self.current_energy = self.new_energy
-        return system_changes
 
-    def _move_rejected(self, system_changes: SystemChanges) -> SystemChanges:
-        logger.debug('Move rejected, undoing system changes: %s', system_changes)
+    def _move_rejected(self, system_changes: SystemChanges) -> None:
+        logger.debug("Move rejected, undoing system changes: %s", system_changes)
         self.generator.on_move_rejected(system_changes)
 
-        # Move rejected, no changes are made
-        system_changes = [
-            SystemChange(index=change.index,
-                         old_symb=change.old_symb,
-                         new_symb=change.old_symb,
-                         name=change.name) for change in system_changes
-        ]
-        logger.debug('Reversed system changes: %s', system_changes)
-        return system_changes
-
     def count_atoms(self) -> Dict[str, int]:
         """Count the number of each element."""
         return dict(Counter(self.atoms.symbols))
 
-    def _mc_step(self) -> Tuple[float, bool]:
+    def _mc_step(self) -> MCStep:
         """Make one Monte Carlo step by swithing two atoms."""
         self.current_step += 1
-
         system_changes = self.generator.get_trial_move()
         self.trial_move = system_changes
 
@@ -320,16 +354,39 @@ class Montecarlo(BaseMC):
         move_accepted = self._calculate_step(system_changes)
 
         updater = self._move_accepted if move_accepted else self._move_rejected
-        system_changes = updater(system_changes)
+        updater(system_changes)
 
+        step = MCStep(self.current_step, self.current_energy, move_accepted, system_changes)
         # Execute all observers
-        self.execute_observers(system_changes)
-        self.filter.add(self.current_energy)
-        return self.current_energy, move_accepted
+        self.execute_observers(step)
 
-    def execute_observers(self, system_changes: SystemChanges):
+        return step
+
+    def execute_observers(self, last_step: MCStep):
         for interval, obs in self.observers:
             if self.current_step % interval == 0:
-                logger.debug('Executing observer %s at step %d with interval %d.', obs.name,
-                             self.current_step, interval)
-                obs(system_changes)
+                logger.debug(
+                    "Executing observer %s at step %d with interval %d.",
+                    obs.name,
+                    self.current_step,
+                    interval,
+                )
+                obs.observe_step(last_step)
+
+
+def _make_default_swap_generator(evaluator: MCEvaluator) -> RandomSwap:
+    """Construct the default RandomSwap trial move generator.
+    If the evaluator object is not a Cluster Expansion evaluator,
+    then any swaps are allowed.
+
+    If the evaluator is a Cluster Expansion MC evaluator object,
+    then only non-background sites are considered for random swaps.
+    This does *not* constrain swaps between sublattices.
+    """
+    atoms = evaluator.atoms
+    if not isinstance(evaluator, CEMCEvaluator):
+        # Running a non-CE Evaluator
+        return RandomSwap(atoms)
+    # We're a CE evaluator, constrain the background sites
+    non_bkg_indices = evaluator.settings.non_background_indices
+    return RandomSwap(atoms, indices=non_bkg_indices)
